; LPTINFO
; -------------------------------------------------------------------------------
; The following code is based on the IEEE-1284 probe code from Linux, which was
; written by Phil Blundell, Carsten Gross and Jose Renau and on Turbo pascal
; translation by unknown author.
; Implementation in assembler (TASM 4.1) is written by Laaca
; I am not sure about timing of signals because original source claims that it is
; important but it was difficult to rewrite in assembler preciously.

; This program can be spread under conditions of GNU/GPL licence.

			; problem with "outb baseport, ..." is that it does
			; print the sent character on non-ieee1284 printers!
%define write4 0	; set to 1 to "outb baseport, 4" to init nibble mode

; --- .8086
; --- SEGMENT	CODE	CLASS=CODE	ALIGN=2
	org 100h

; --- STARTUPCODE
	jmp @start

q 	db 0
y	db 0
; --- z	db 0	; unused?
f	db 0	; "which nibble are we in", toggles between 0 and 1
quit	db 0
baseport dw 0

warr	db 'This program is free software; you can redistribute it and/or modify',13,10
	db 'it under the terms of the GNU General Public License as published by',13,10
	db 'the Free Software Foundation; either version 2 of the License, or (at',13,10
	db 'your option) any later version.',13,10,'$'

help	db 'LPTinfo v1.3   (programmed by Laaca, july 2007)',13,10
	db 'Writes PnP identification string from device on parallel port.',13,10,13,10
	db 'LPTTEST [/1 | /2 | /3 | /W]',13,10
	db '/1 - /3     prints info about device on LPT1 - LPT3',13,10
	db '/W          displays warranty information',13,10,'$'

ieee		db 'LPTinfo v1.3',13,10,'IEEE-1284 Parallel port device scan.',13,10,13,10,'$'
present		db 'IEEE-1284 compatible device found.',13,10,'$'
second		db 'Device ID available.',13,10,'$'
read_1_timeout	db 'Read1 timeout!',13,10,'$'
read_2_timeout	db 'Read2 timeout!',13,10,'$'
ok_id		db 'ID Read finished OK',13,10,13,10,'$'
err_id		db 'ID Read terminated with error!',13,10,13,10,'$'
err_invalid	db 'ID Read error - invalid data',13,10,'$'
our_size	db 'Size received  : $'
ret_size	db 'Size announced : $'
inf_ret		db 'Info returned  : $'
formatter	db 13,10,'                 $'
failed_0	db ' is not installed.',13,10,'$'
failed_1	db 'No IEEE-1284 compatible device found.',13,10,'$'
failed_2	db 'Device ID not readable.',13,10,'$'
low_ram		db 'Not enough RAM free',13,10,'$'
crlf		db 13,10,'$'
devstr1		db 'Device on port LPT'
devstr2		db '? has address: '
devstr3		db '????h',13,10,'$'
number		db '????$'	; for int2str

numbers	db '0123456789ABCDEF'
calib	dw 0,0,0	; loop iterations per millisecond
devnum	dw 1		; printer number: 1, 2 or 3

read_baseport:		; set baseport to baseport of printer devnum
	push es
	push di
	xor di,di
	mov es,di
	mov ax,[devnum]
	add ax,ax
	mov di,406h	; offset of an array of 3 words
	add di,ax	; for LPT1 will be result 408h
	mov ax,[es:di]
	mov [baseport],ax
	pop di
	pop es
	retn


delay:			; wait 1 msec (old version: ah msec)
	push ax
	push bx
	push cx
	push dx

	mov ax,[calib]
	mov bx,[calib+2]
	mov cx,[calib+4]
	mov dx,[baseport]
@de1:	push ax
	in al,dx	; dummy read from port - like in calibration routine
	pop ax
	sub ax,1
	sbb bx,0
	sbb cx,0
	jnz @de1
	pop dx
	pop cx
	pop bx
	pop ax
	retn


XWait:				; AH=number of loops
	push dx
	push bx
	mov bl,1
	mov dx,[baseport]
	inc dx	; DX=379h
@xcycle:
	in al,dx
	inc bl
	call delay
	cmp bl,40
	jz @quit_cycle
	and al,ch
	cmp al,cl
	jnz @xcycle
@quit_cycle:
	mov al,1
	cmp bl,40
	jb @is_true
	xor al,al
@is_true:
	pop bx
	pop dx
	retn

read_nibble:	; read 4 bits of info
	push dx
	mov dx,[baseport]
	inc dx	; DX=379h
	in al,dx
	shr al,1
	shr al,1
	shr al,1
	and al,0f7h
	test al,10h
	jnz @not_zero
	or al,8
@not_zero:
	and al,0fh
	pop dx
	retn

print:	; string must be in DS:SI and ended by $
	push dx
	mov dx,si
	mov ah,9
	int 21h
	pop dx
	retn

hex2str:	; BX = value, write unterminated 4 char string to DS:SI
	push cx	; modifies AX SI (BX ends up the same after the loop)
	mov cx,0404h	; shift and count
@hex1:	rol bx,cl
	push bx
	and bx,15	; one digit
	mov al,[numbers+bx]
	mov [ds:si],al
	inc si
	pop bx
	dec ch
	jnz @hex1
	sub si,4
	mov cx,3
	mov al,' '	; now replace leading zeros by spaces
@hex2:
	cmp byte [ds:si],'0'
	jnz @hex3
	mov [ds:si],al
	inc si
	loop @hex2
@hex3:
	pop cx
	retn

int2str:	; BX = value, show as 4 char string, modifies AX BX
	push si
	push cx
	push dx
	cmp bx,9999
	jbe @int2
	mov bx,9999	; limit value
@int2:	mov ax,bx
	xor dx,dx
	mov bx,100
	mov cl,4
	div bx		; now dx is low digits, ax is high ones
	xor bx,bx
	aam		; ah=al/10 al=al mod 10
	shl ah,cl
	or bh,ah	; highest digit
	or bh,al	; next digit
	mov ax,dx
	aam		; ah=al/10 al=al mod 10
	shl ah,cl
	or bl,ah	; next digit
	or bl,al	; BX is now the BCD form of the value	
	mov si,number	; offset
	push si
	call hex2str
	pop si
	call print
	pop dx
	pop cx
	pop si
	retn


set_delay:
	push es
	push di
	push ax
	push bx
	push cx
	push dx
	push bp

	xor ax,ax
	mov es,ax
	mov di,46ch
	mov bp,[es:di]
	mov dx,[baseport]
	xor ax,ax
	xor bx,bx
	xor cx,cx
	xor si,si
@st1:
	cmp bp,[es:di]
	jz @st1
	mov bp,[es:di]	; start of a new tick
	xor di,di	; tick count

@st2:	push ax
	in al,dx	; dummy read from port - reduces number of cycles
	pop ax
	add ax,1
	adc bx,0
	adc cx,0
	cmp bp,[es:46ch]
	jz @st2
	mov bp,[es:46ch]
	inc di
	cmp di,9	; 9 ticks are roughly 512 msec ;-)
	jnz @st2

@st3:	shr cx,1	; shift right everything by 9 bits
	rcr bx,1
	rcr ax,1
	dec di
	jnz @st3

	mov [calib],ax
	mov [calib+2],bx
	mov [calib+4],cx
	pop bp
	pop dx
	pop cx
	pop bx
	pop ax
	pop di
	pop es
	retn


GetCommandline:
	mov si,80h
	mov cl,[si]	; Length of command line
	cmp cl,1
	ja @gcl_go	; empty command line?
	jmp @gcl_nothing
@gcl_go:
	mov bh,0
	add si,2	; Start of command line
@gcl:
	mov al,[si+bx]
	cmp al,'-'
	jz @gcl_switch
	cmp al,'/'
	jz @gcl_switch
	cmp al,' '
	jnz @help_message
	mov ah,2
	int 21h
	inc bl
	cmp bl,cl
	jnz @gcl
	jmp @gcl_nothing	; all command line was only spaces

@gcl_switch:
	mov al,[si+bx+1]
	cmp al,'a'
	jb @dont_convert
	cmp al,'z'
	ja @dont_convert
	sub al,32	; convert to upper case
@dont_convert:
	cmp al,'W'
	jz @warranty_message
	cmp al,'1'
	jb @help_message
	cmp al,'3'
	ja @help_message
	sub al,'0'
	mov [devnum],al
	jmp @gcl_nothing

@warranty_message:
	lea si,[warr]
	call print
	mov ax,4c00h
	int 21h


@help_message:
	lea si,[help]
	call print
	mov ax,4c00h
	int 21h

@gcl_nothing:
	retn

; ==========================================================================
; ==========================================================================

@start:
	call GetCommandline
	lea si,[ieee]
	call print

	call read_baseport

	lea si,[devstr1]
	mov bx,[devnum]
	add bl,'0'	; devnum only has 1 digit
	mov byte [devstr2],bl
	cmp WORD [baseport],0
	jnz @port_exists
	mov byte [devstr2+1],'$'	; terminate string
	call print
	lea si,[failed_0]	; no such port
	call print
	jmp @go_exit
@port_exists:
	mov bx,[baseport]
	lea si,[devstr3]
	call hex2str
	lea si,[devstr1]
	call print

	call set_delay

	mov di,pnpdata	; offset
	mov ax,sp
	sub ax,2
	sub ax,di
	jnc enough_ram
	lea si,[low_ram]
	call print
	mov ax,4c01h
	int 21h
enough_ram:

	mov dx,[baseport]	; DX=378h
	mov al,4
%if write4
; *** insert some handshaking here
	out dx,al	; Most of devices don't need it but some yes
%endif
	call delay

	add dx,2	; DX=37ah
	in al,dx
	and al,0f7h
	out dx,al
	in al,dx
	or al,2
	out dx,al

	mov cx,7838h
	call XWait
	cmp al,0
	jnz @okay1
	jmp @no_first

@okay1:
; ------------------------------------------------------------------------
	lea si,[present]
	call print
	in al,dx
	or al,1
	out dx,al	; strobe high
	in al,dx
	and al,0fdh
	out dx,al	; autofeed low
	call delay
	in al,dx
	and al,0feh
	out dx,al	;  strobe low
	call delay

	mov cx,2000h
	call XWait
	cmp al,0
	jnz @okay2
	jmp @no_second
@okay2:

; ------------------------------------------------------------------------

	lea si,[second]
	call print

	mov byte [f],0
	mov BYTE [quit],0
	mov bx,0	; BX=index into array
	call delay

@repeat_cycle:
	in al,dx	; autofeed high
	or al,2
	out dx,al

	mov cx,4000h
	call XWait

	cmp al,0
	jnz @no_timeout1

	; ---------------------------------------------------------------------
	lea si,[read_1_timeout]
	call print
	in al,dx	; autofeed low
	and al,0fdh
	out dx,al
	mov BYTE [quit],1
	; ---------------------------------------------------------------------
@no_timeout1:

	call read_nibble	; get 4 bits
	mov [y],al
	in al,dx
	and al,0fdh
	out dx,al	; autofeed low
	mov cx,4040h
	call XWait
	cmp al,0
	jnz @no_timeout2
	; ---------------------------------------------------------------------
	lea si,[read_2_timeout]
	call print
	mov BYTE [quit],1
	; ---------------------------------------------------------------------
@no_timeout2:
	cmp byte [f],0	; f signals when a byte is
	jz @f_not_set	; ready - each 2 passes
	xor byte [f],1

	mov al,[y]
	mov cl,4
	shl al,cl
	add al,[q]
	mov [q],al

	mov al,[q]
	mov [es:di+bx],al
	inc bx

	dec dx	; DX=379h
	in al,dx	; eof?
	inc dx	; DX=37ah
	and al,8
	cmp al,8
	jnz @no_eof
	mov BYTE [quit],1
@no_eof:
	jmp @f_join

@f_not_set:
	xor byte [f],1
	mov al,[y]
	mov [q],al
@f_join:

	cmp bx,1023
	jna @s_no_greater	; out of table space ?
	lea si,[err_invalid]
	call print
	jmp @go_exit
@s_no_greater:


	cmp BYTE [quit],0
	jnz @dont_repeat
	jmp @repeat_cycle
@dont_repeat:
; -----------------------------------------------------------------------

	in al,dx
	and al,0fdh
	or al,8
	out dx,al	; terminate read

	mov cx,8000h
	call XWait

	cmp al,0
	jz @id_error
	lea si,[ok_id]
	call print
	jmp @id_join
@id_error:
	lea si,[err_id]
	call print
@id_join:

; Display Result, nicely formatted

	lea si,[our_size]
	call print
	call int2str	; print bx
	lea si,[crlf]
	call print

	lea si,[ret_size]
	call print
	mov ah,[es:di]
	mov al,[es:di+1]
	mov cx,bx
	mov bx,ax
	call int2str	; print bx
	lea si,[crlf]
	call print
	lea si,[inf_ret]
	call print

	mov bx,2
@write_cycle:
	mov al,[es:di+bx]
	cmp al,59
	jnz @normal_char
	lea si,[formatter]
	call print
	jmp @write_join
@normal_char:
	mov dl,al
	mov ah,2
	int 21h
@write_join:
	inc bx
	cmp bx,cx
	jb @write_cycle
	lea si,[crlf]
	call print
	lea si,[crlf]
	call print
	jmp @go_exit

@no_second:
	lea si,[failed_2]
	call print
	jmp @go_exit

@no_first:
	lea si,[failed_1]
	call print
	mov dx,[baseport]
	add dx,2
	in al,dx
	and al,0fdh
	or al,8
	out dx,al

@go_exit:
	xor bx,bx
	mov ax,4c00h
	int 21h

pnpdata:	; label, buffer for received data starts here
	db 0
; --- GROUP	DGROUP	CODE
; --- end

